(ns com.aurellem.gb.pokemon
  (:use (com.aurellem.gb gb-driver util constants characters
                         moves types items status dv species
                         experience stats pokemon-presets
                         ))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(def pokemon-names-start 0xD2B4)

(defn set-party-number
  ([^SaveState state new-party-num]
     (-> state
         (set-memory 0xD162 new-party-num)
         (set-memory (+ 0xD162 new-party-num 1) 0xFF)))
  ([new-party-num]
     (set-party-number @current-state new-party-num)))

(def party-number-address 0xD162)

(defn party-number
  ([^SaveState state]
     (aget (memory state) party-number-address))
  ([] (party-number @current-state)))

(defn party-names
  ([^SaveState state]
     (let [raw-names
           (subvec (vec (memory state))
                   pokemon-names-start
                   (+ pokemon-names-start
                      (* name-width 6)))]
       (map
        read-name
        (take
         (party-number state)
          (partition name-width
                   raw-names)))))
  ([] (party-names @current-state)))

(defn pokemon-nickname
  ([^SaveState state poke-num]
     (nth (party-names state) poke-num))
  ([poke-num]
     (pokemon-nickname @current-state poke-num)))

(defn rename-pokemon
  ([^SaveState state n new-name]
     (assert (<= 0 n (dec (party-number state))))
     (assert (<= (count new-name) max-name-length))
     (set-memory-range
      state
      (+ (* n name-width) pokemon-names-start)
      (concat (str->character-codes new-name) [end-of-name-marker])))
  ([n new-name]
     (rename-pokemon @current-state n new-name)))

(def OT-start 0xD272)

(defn original-trainers
  ([^SaveState state]
     (let [raw-names
           (subvec (vec (memory state))
                   OT-start
                   (+ OT-start
                      (* name-width 6)))]
       (map read-name
            (take (party-number state)
                  (partition name-width raw-names)))))
  ([] (original-trainers @current-state)))

(defn read-OT-name
  ([^SaveState state poke-num]
     (nth (original-trainers state) poke-num))
  ([poke-num] (read-OT-name @current-state poke-num)))

(defn set-OT-name
  "Set the OT name for a pokemon.
   Note that a pokemon is still considered 'yours' if
   the OT ID is the same as your own."
  ([^SaveState state poke-num new-name]
     (assert (<= 0 poke-num (dec (party-number state))))
     (assert (<= (count new-name) max-name-length))
     (set-memory-range
      state
      (+ (* poke-num name-width) OT-start)
      (concat (str->character-codes new-name) [end-of-name-marker])))
  ([n new-name]
     (set-OT-name @current-state n new-name)))

(def OT-ID-addresses [0xD176 0xD1A2 0xD1CE 0xD1FA 0xD226 0xD252])

(defn read-OT-id
  ([^SaveState state poke-num]
     (let [mem (memory state)
           start (OT-ID-addresses poke-num)]
     (glue-bytes
      (aget mem start)
      (aget mem (inc start)))))
  ([poke-num] (read-OT-id @current-state poke-num)))

(defn set-OT-id
  ([^SaveState state poke-num new-OT-num]
     (assert (<= 0 poke-num 5))
     (assert (<= 0 new-OT-num 0xFFFF))
     (set-memory-range 
      state
      (OT-ID-addresses poke-num)
      (disect-bytes-2 new-OT-num)))
  ([poke-num new-OT-num]
     (set-OT-id @current-state poke-num new-OT-num)))

(def unknown "[[[UNKNOWN]]]")

(def unknown "")

(def pokemon-1-record
  {0xD16A "Color Map"            ;; 0
   0xD16B "Current-HP (h)"       ;; 1
   0xD16C "Current-HP (l)"       ;; 2
   0XD16D "Unused"               ;; 3
   0xD16E "Status"               ;; 4
   0xD16F "Type 1"               ;; 5
   0xD170 "Type 2"               ;; 6
   0xD171 "scratch/C.R."         ;; 7
   0xD172 "Move 1 ID"            ;; 8
   0xD173 "Move 2 ID"            ;; 9
   0xD174 "Move 3 ID"            ;; 10
   0xD175 "Move 4 ID"            ;; 11
   0xD176 "OT-ID (h)"            ;; 12
   0xD177 "OT-ID (l)"            ;; 13
   0xD178 "Exp. Points (h)"      ;; 14
   0xD179 "Exp. Points (m)"      ;; 15
   0xD17A "Exp. Points (l)"      ;; 16
   0xD17B "HP Exp. (h)"          ;; 17
   0xD17C "HP Exp. (l)"          ;; 18
   0xD17D "Attack Exp. (h)"      ;; 19
   0xD17E "Attack Exp. (l)"      ;; 20
   0xD17F "Defense Exp. (h)"     ;; 21
   0xD180 "Defense Exp. (l)"     ;; 22
   0xD181 "Speed Exp. (h)"       ;; 23
   0xD182 "Speed Exp. (l)"       ;; 24
   0xD183 "Special Exp. (h)"     ;; 25
   0xD184 "Special Exp. (l)"     ;; 26
   0xD185 "DV Atk/Def"           ;; 27
   0xD186 "DV Speed/Spc"         ;; 28
   0xD187 "PP Move 1"            ;; 29
   0xD188 "PP Move 2"            ;; 30
   0xD189 "PP Move 3"            ;; 31
   0xD18A "PP Move 4"            ;; 32
   0xD18B "Current Level"        ;; 33
   0xD18C "HP Total (h)"         ;; 34
   0xD18D "HP Total (l)"         ;; 35
   0xD18E "Attack   (h)"         ;; 36
   0xD18F "Attack   (l)"         ;; 37
   0xD190 "Defense  (h)"         ;; 38
   0xD191 "Defense  (l)"         ;; 39
   0xD192 "Speed    (h)"         ;; 40
   0xD193 "Speed    (l)"         ;; 41
   0xD194 "Special  (h)"         ;; 42
   0xD195 "Special  (l)"         ;; 43
   })

(defn pokemon-record
  ([^SaveState state pokemon-num]
     (assert (<= 0 pokemon-num 5))
     (let [base (+ (* pokemon-num  pokemon-record-width) 0xD16A)]
       (subvec (vec (memory state)) base
               (+ base pokemon-record-width))))
  ([pokemon-num] (pokemon-record @current-state pokemon-num)))

(defn set-pokemon-record
  ([^SaveState state pokemon-num new-data]
     (assert (<= 0 pokemon-num 5))
     (let [base (+ (* pokemon-num  pokemon-record-width) 0xD16A)]
       (set-memory-range state base new-data)))
  ([pokemon-num new-data]
     (set-pokemon-record @current-state pokemon-num new-data)))

(defn print-pokemon-record
  ([^SaveState state pokemon-num]
     (assert (<= 0 pokemon-num 5))
     (let [poke-data (pokemon-record state pokemon-num)
           backbone (sort (keys pokemon-1-record))]
       (println "Pokemon " pokemon-num  " -- "
                (nth (party-names state)
                     pokemon-num) \newline)

       (println "    Desc.          | Hex  | Dec |  Binary  |")
       (println "-------------------+------+-----+----------+")
       (dorun
        (map
         (comp println
               (fn [desc data]
                 (format "%-16s   | 0x%02X | %3d | %s |"
                         desc data data
                         (let [s (Integer/toBinaryString data)]
                           (apply
                            str 
                            (concat (repeat (- 8 (count s)) "0" )
                                    s))))))
         (map pokemon-1-record backbone)
         poke-data))))
  ([pokemon-num]
     (print-pokemon-record @current-state pokemon-num)))

(defn pokemon
  ([^SaveState state poke-num]
     (assert (<= 0 poke-num 5))
     (let [dv-values (read-DV state poke-num)
           type (read-type state poke-num)
           species (read-species state poke-num)
           species2 (read-species2 state poke-num)
           moves (read-moves state poke-num)
           moves-pp (mapv (partial read-pp state
                                   poke-num)
                          (range (count moves)))
           nickname (pokemon-nickname state poke-num)
           status (read-status state poke-num)
           stats (read-stats state poke-num)
           experience (read-experience state poke-num)
           OT-name (read-OT-name state poke-num)
           ID (read-OT-id state poke-num)]
       {;; persistent 
        :name nickname
        :species species
        :species2 species2
        :type type
        :dv dv-values
        :original-trainer OT-name
        :ID ID
        :moves (mapv vector moves moves-pp)

        ;; ephemerial
        :status status
        :stats stats
        :experience experience
        }))
  ([poke-num]
     (pokemon @current-state poke-num)))

(def status-message
  {:sleep-6 "sleeping. It will wake in six turns."
   :sleep-5 "sleeping. It will wake in five turns."
   :sleep-4 "sleeping. It will wake in four turns."
   :sleep-3 "sleeping. It will wake in three turns."
   :sleep-2 "sleeping. It will wake in two turns."
   :sleep-1 "sleeping. It will wake in one turn."
   :poisoned "poisoned."
   :frozen "frozen solid."
   :burned "burned."
   :paralyzed "paralyzed."})

   
(defn print-pokemon
  ([^SaveState state poke-num]
     (let [info (pokemon state poke-num)]
       (printf
        (str
         "##################################"
         "##################################\n"
         "#                                 "
         "                                 #\n"
         "#  %-44s"
         "%-20s#\n"
         "#                                 "
         "                                 #\n"
         "##################################"
         "##################################\n\n")
        
        (str
         (:name info)
         (str
          " [" (.toUpperCase
                (.substring (str (:species info)) 1)) "]")
         (str "  Lvl." (format "%-3d" (:level (:stats info)))))
        (str (:original-trainer info) " / " (:ID info)))
       
       (println
        (str 
         (str "-----------------------------------"
              "---------------------------------\n" )
         (str "| Stats     |    HP    |  Attack  "
              "| Defense  |  Speed   | Special  |\n")
         (str "+-----------+----------+----------"
              "+----------+----------+----------+")))

       (printf
        (str "|%-11s|   %5d  |   %5d  "
             "|   %5d  |   %5d  |   %5d  |\n")
        "DV Values" (:hp (:dv info)) (:attack (:dv info))
        (:defense (:dv info)) (:speed (:dv info))
        (:special (:dv info)))

       (let [c (:stats info)]
         (printf
          (str "|%-11s|%8s  |   %5d  "
               "|   %5d  |   %5d  |   %5d  |\n")
          "Current" (str (:current-hp c) "/" (:hp c)) (:attack c)
          (:defense c) (:speed c)
          (:special c)))
              
       (let [e (:experience info)]
         (printf
          (str "|%-11s|   %5d  |   %5d  "
               "|   %5d  |   %5d  |   %5d  |\n")
          "Experience" (:hp-exp e) (:attack-exp e)
          (:defense-exp e) (:speed-exp e)
          (:special-exp e)))
       (println
        (str "+-----------+----------+----------"
             "+----------+----------+----------+"))

       (print "\n")
       (println "+------------------+----+--------+--------+")
       (println "|       Move       | PP | Max PP | PP UPs |")
       (println "+------------------+----+--------+--------+")

       (dorun
        (for [[name {:keys [pp-ups current-pp]}] (:moves info)]
          (printf
           "| %-17s| %2d |   %02d   |   %02d   |\n"
           (.substring (str name) 1)
           current-pp (max-pp name pp-ups) pp-ups)))

       (println "+------------------+----+--------+--------+\n")
       
       (println "Total Experience:" (:main-exp (:experience info)))
       (if (not= :normal (:status info))
         (println "\n* This pokemon is currently"
                  (status-message (:status info))))
       (if (not= (:species info) (:species2 info))
         (println "\n* This pokemon has a secondary species"
                  (str
                   "("
                   (.substring (str (:species2 info)) 1) ")\n")
                  "  that does not match its primary species."))
       (if (not= (:type info) (pokemon->type (:species info)))
         (println "\n* This pokemon has a type"
                  (str (:type info))
                  "which \n"
                  "  is inconsistent with the default type for its"
                  "species."))))
  ([poke-num]
     (print-pokemon @current-state poke-num)))

(defn print-team [] 
  (dorun (map print-pokemon (range (party-number)))))


(defn give-status-all
  ([^SaveState state status]
     (reduce (fn [state num]
               (give-status state num status))
             state
             (range (party-number state))))
  ([status]
     (give-status-all @current-state status)))

(defn expand-pokemon
  "Given a map describing a pokemon, fill in any missing
   values based on the ones already present."
  [pokemon]
  (-> (merge pokemon-base pokemon)
      ;; if no nickname is supplied, default to the
      ;; uppercase name of the species, as r/b/y do
      ;; when a pokemon is captured.
      ((fn [pokemon]
         (if (nil? (:name pokemon))
           (assoc pokemon :name (.toUpperCase
                                 (.substring
                                  (str (:species pokemon)) 1)))
           pokemon)))
      ;; species2 should almost always just be the
      ;; same as species.
      ((fn [pokemon]
         (if (nil? (:species2 pokemon))
           (assoc pokemon :species2 (:species pokemon))
           pokemon)))

      ;; enable the date in :moves to be any combo of 
      ;; [:move-1 :move-2]
      ;; [[:move-1 {:pp 20}] :move-2]
      ;; [[:move-1 {:pp 20 :pp-up 3}] :move-2]
      ;; default to full pp for the move, with no
      ;; pp-ups.
      ((fn [pokemon]
         (let [moves (:moves pokemon)]
           (assoc pokemon :moves
                  (for [move moves]
                    (cond
                      (keyword? move)
                      [move {:current-pp (max-pp move 3) :pp-ups 3}]
                      (vector? move)
                      [(first move)
                       (merge {:current-pp (max-pp (first move) 3)
                               :pp-ups 3} (second move))]))))))
      ;; The game stores the pokemon's type redundantly
      ;; along with the species.  If it's not specified
      ;; then it should default to that species default type.
      ((fn [pokemon]
         (if (nil? (:type pokemon))
           (assoc pokemon :type
                  (pokemon->type (:species pokemon)))
           pokemon)))))

(def mint-berry-item-code-gsc 0x54)

(defn give-pokemon
  ([^SaveState state poke-num pokemon]
     (let [pokemon* (expand-pokemon pokemon)]
       (-> state
           ;; expand roster if necessary
           ((fn [state]
              (if (< (dec (party-number state)) poke-num)
                (set-party-number state (inc poke-num)) state)))
           (rename-pokemon poke-num (:name pokemon*))
           (give-DV poke-num (:dv pokemon*))
           (give-type poke-num (:type pokemon*))
           (set-species poke-num (:species pokemon*))
           (set-species2 poke-num (:species2 pokemon*))
           (set-OT-id poke-num (:ID pokemon*))
           (set-OT-name poke-num (:original-trainer pokemon*))
           (give-moves-pps poke-num (:moves pokemon*))
           (give-status poke-num (:status pokemon*))
           (give-stats  poke-num (:stats  pokemon*))
           (give-experience poke-num (:experience pokemon*))
           (set-memory (+ 0xD16D
                          (* pokemon-record-width poke-num))
                          mint-berry-item-code-gsc))))
  ([poke-num pokemon]
     (give-pokemon @current-state poke-num pokemon)))

(defn edit-pokemon
  ([^SaveState state poke-num new-pokemon-data]
     (give-pokemon state poke-num
                   (merge (pokemon state poke-num)
                          new-pokemon-data)))
  ([poke-num new-pokemon-data]
     (edit-pokemon @current-state poke-num new-pokemon-data)))

(defn give-powerful-party
  ([^SaveState state]
     (-> state
          (give-pokemon 0 tauros)
          (give-pokemon 1 chansey)
          (give-pokemon 2 alakazam)
          (give-pokemon 3 golem)
          (give-pokemon 4 exeggutor)
          (give-pokemon 5 starmie)

          ))
  ([] (give-powerful-party @current-state)))